"""MIP Model

This module implements abstractions for working with Mixed-Integer Programming
Models.

The Documentation for Python-MIP is available at:
https://python-mip.readthedocs.io/en/latest/

A PDF version is also available:
https://media.readthedocs.org/pdf/python-mip/latest/python-mip.pdf

"""

from typing import List, Tuple

from mip.constants import *
from math import inf
from builtins import property
from os import environ


class Column:
    """ A column (variable) in the constraint matrix

        To create a column see Model.add_var

    """

    def __init__(self,
                 constrs: List["Constr"] = None,
                 coeffs: List[float] = None):
        self.constrs = constrs
        self.coeffs = coeffs


class Constr:
    """ A row (constraint) in the constraint matrix

        a constraint can be added to the model using the overloaded operator
        +=, e.g., if m is a model:

        m += 3*x1 + 4*x2 <= 5

        summation expressions are also supported:

        m += xsum(x[i] for i in range(n)) == 1

    """

    def __init__(self, model: "Model", idx: int, name: str = ""):
        self.model = model
        self.idx = idx
        self.name = name  # discuss this var

    def __hash__(self) -> int:
        return self.idx

    def __str__(self) -> str:
        return self.name

    @property
    def pi(self) -> float:
        return self.model.solver.constr_get_pi(self)

    @property
    def expr(self) -> "LinExpr":
        return self.model.solver.constr_get_expr(self)

    @expr.setter
    def expr(self, value: "LinExpr") -> None:
        self.model.solver.constr_set_expr(self, value)


class LinExpr:
    """ A Linear Expression

    Linear expressions are used to enter the objective function and the model \
    constraints.

    Consider a model object m, the objective function of m can be specified as:

    m += 10*x1 + 7*x4

    summation can also be used:

    m += xsum(3*x[i] i in range(n)) - xsum(x[i] i in range(m))

    If not specified in the construction of the model object, it is assumed
    that the model is a minimization one.

    A constraint is just a linear expression with the addition of a sense (==,
    <= or >=) and a right hand side, e.g.:

    m += x1 + x2 + x3 == 1

    """

    def __init__(self,
                 variables: List["Var"] = None,
                 coeffs: List[float] = None,
                 const: float = 0,
                 sense: str = ""):
        self.const = const
        self.expr = {}
        self.sense = sense

        if variables:
            assert len(variables) == len(coeffs)
            for i in range(len(coeffs)):
                if coeffs[i] == 0:
                    continue
                self.add_var(variables[i], coeffs[i])

    def __add__(self, other) -> "LinExpr":
        result = self.copy()
        if isinstance(other, Var):
            result.add_var(other, 1)
        elif isinstance(other, LinExpr):
            result.add_expr(other)
        elif isinstance(other, (int, float)):
            result.add_const(other)
        return result

    def __radd__(self, other) -> "LinExpr":
        return self.__add__(other)

    def __iadd__(self, other) -> "LinExpr":
        if isinstance(other, Var):
            self.add_var(other, 1)
        elif isinstance(other, LinExpr):
            self.add_expr(other)
        elif isinstance(other, (int, float)):
            self.add_const(other)
        return self

    def __sub__(self, other) -> "LinExpr":
        result = self.copy()
        if isinstance(other, Var):
            result.add_var(other, -1)
        elif isinstance(other, LinExpr):
            result.add_expr(other, -1)
        elif isinstance(other, (int, float)):
            result.add_const(-other)
        return result

    def __rsub__(self, other) -> "LinExpr":
        return (-self).__add__(other)

    def __isub__(self, other) -> "LinExpr":
        if isinstance(other, Var):
            self.add_var(other, -1)
        elif isinstance(other, LinExpr):
            self.add_expr(other, -1)
        elif isinstance(other, (int, float)):
            self.add_const(-other)
        return self

    def __mul__(self, other) -> "LinExpr":
        assert isinstance(other, int) or isinstance(other, float)
        result = self.copy()
        result.const *= other
        for var in result.expr.keys():
            result.expr[var] *= other

        # if constraint sense will change
        if self.sense == GREATER_OR_EQUAL and other <= -1e-8:
            self.sense = LESS_OR_EQUAL
        if self.sense == LESS_OR_EQUAL and other <= -1e-8:
            self.sense = GREATER_OR_EQUAL

        return result

    def __rmul__(self, other) -> "LinExpr":
        return self.__mul__(other)

    def __imul__(self, other) -> "LinExpr":
        assert isinstance(other, int) or isinstance(other, float)
        self.const *= other
        for var in self.expr.keys():
            self.expr[var] *= other
        return self

    def __truediv__(self, other) -> "LinExpr":
        assert isinstance(other, int) or isinstance(other, float)
        result = self.copy()
        result.const /= other
        for var in result.expr.keys():
            result.expr[var] /= other
        return result

    def __itruediv__(self, other) -> "LinExpr":
        assert isinstance(other, int) or isinstance(other, float)
        self.const /= other
        for var in self.expr.keys():
            self.expr[var] /= other
        return self

    def __neg__(self) -> "LinExpr":
        return self.__mul__(-1)

    def __str__(self) -> str:
        result = []

        if self.expr:
            for var, coeff in self.expr.items():
                result.append("+ " if coeff >= 0 else "- ")
                result.append(str(abs(coeff)) if abs(coeff) != 1 else "")
                result.append("{var} ".format(**locals()))

        if self.sense:
            result.append(self.sense + "= ")
            result.append(str(abs(self.const)) if self.const < 0 else "- " +
                                                                      str(abs(self.const)))
        elif self.const != 0:
            result.append(
                "+ " + str(abs(self.const)) if self.const > 0 else "- " +
                                                                   str(abs(self.const)))

        return "".join(result)

    def __eq__(self, other) -> "LinExpr":
        result = self - other
        result.sense = "="
        return result

    def __le__(self, other) -> "LinExpr":
        result = self - other
        result.sense = "<"
        return result

    def __ge__(self, other) -> "LinExpr":
        result = self - other
        result.sense = ">"
        return result

    def add_const(self, const: float) -> None:
        self.const += const

    def add_expr(self, expr: "LinExpr", coeff: float = 1) -> None:
        self.const += expr.const * coeff
        for var, coeff_var in expr.expr.items():
            self.add_var(var, coeff_var * coeff)

    def add_term(self, expr, coeff: float = 1) -> None:
        if isinstance(expr, Var):
            self.add_var(expr, coeff)
        elif isinstance(expr, LinExpr):
            self.add_expr(expr, coeff)
        elif isinstance(expr, float) or isinstance(expr, int):
            self.add_const(expr)

    def add_var(self, var: "Var", coeff: float = 1) -> None:
        if var in self.expr:
            if -EPS <= self.expr[var] + coeff <= EPS:
                del self.expr[var]
            else:
                self.expr[var] += coeff
        else:
            self.expr[var] = coeff

    def copy(self) -> "LinExpr":
        copy = LinExpr()
        copy.const = self.const
        copy.expr = self.expr.copy()
        copy.sense = self.sense
        return copy


class Model:
    """ Mixed Integer Programming Model

    This is the main class, providing methods for building, optimizing,
    querying optimization results and reoptimizing Mixed-Integer Programming
    Models.

    To check how models are created please see the examples included.

    """

    def __init__(self, name: str = "",
                 sense: str = MINIMIZE,
                 solver_name: str = ""):
        """Model constructor

        Creates a Mixed-Integer Linear Programming Model. The default model
        optimization direction is Minimization. To store and optimize the model
        the MIP package automatically searches and connects in runtime to the
        dynamic library of some MIP solver installed on your computer, nowadays
        gurobi and cbc are supported. This solver is automatically selected,
        but you can force the selection of a specific solver with the parameter
        solver_name.

        Args:
            name (str): model name
            sense (str): MINIMIZATION ("MIN") or MAXIMIZATION ("MAX")
            solver_name: gurobi or cbc, searches for which
                solver is available if not informed

        """
        # initializing variables with default values
        self.name = name
        self.solver_name = solver_name
        self.solver = None
        if "solver_name" in environ:
            solver_name = environ["solver_name"]
        if "solver_name".upper() in environ:
            solver_name = environ["solver_name".upper()]

        self.__mipStart = []

        # list of constraints and variables
        self.constrs = []
        self.constrs_by_name = {}
        self.vars = []
        self.vars_by_name = {}
        self.cut_generators = []

        if solver_name.upper() == GUROBI:
            from mip.gurobi import SolverGurobi
            self.solver = SolverGurobi(self, name, sense)
        elif solver_name.upper() == CBC:
            from mip.cbc import SolverCbc
            self.solver = SolverCbc(self, name, sense)
        else:
            # checking which solvers are available
            from mip import gurobi
            if gurobi.has_gurobi:
                from mip.gurobi import SolverGurobi
                self.solver = SolverGurobi(self, name, sense)
                self.solver_name = GUROBI
            else:
                from mip import cbc
                from mip.cbc import SolverCbc
                self.solver = SolverCbc(self, name, sense)
                self.solver_name = CBC

        self.sense = sense

    def __del__(self):
        if self.solver:
            del self.solver

    def __iadd__(self, other) -> "Model":
        if isinstance(other, LinExpr):
            if len(other.sense) == 0:
                # adding objective function components
                self.objective = other
            else:
                # adding constraint
                self.add_constr(other)
        elif isinstance(other, tuple):
            if isinstance(other[0], LinExpr) and isinstance(other[1], str):
                if len(other[0].sense) == 0:
                    self.objective = other[0]
                else:
                    self.add_constr(other[0], other[1])

        return self

    def add_var(self, name: str = "",
                lb: float = 0.0,
                ub: float = INF,
                obj: float = 0.0,
                var_type: str = CONTINUOUS,
                column: "Column" = None) -> "Var":
        """ Creates a new variable

        Adds a new variable to the model.

        Args:
            name (str): variable name (optional)
            lb (float): variable lower bound, default 0.0
            ub (float): variable upper bound, default infinity
            obj (float): coefficient of this variable in the objective function, default 0
            var_type (str): CONTINUOUS ("C"), BINARY ("B") or INTEGER ("I")
            column (Column): constraints where this variable will appear, necessary \
            only when constraints are already created in the model and a new \
            variable will be created.

        Examples:

            To add a variable x which is continuous and greater or equal to zero to model m::

                x = m.add_var()

            The following code creates a vector of binary variables x[0], ..., x[n-1] to model m::

                x = [m.add_var(type=BINARY) for i in range(n)]


        """
        if var_type == BINARY:
            lb = 0.0
            ub = 1.0
        if len(name.strip()) == 0:
            nc = self.solver.num_cols()
            name = "C{:011d}".format(nc)
        idx = self.solver.add_var(obj, lb, ub, var_type, column, name)
        self.vars.append(Var(self, idx, name))
        self.vars_by_name[name] = self.vars[-1]
        return self.vars[-1]

    def add_constr(self, lin_expr: "LinExpr", name: str = "") -> Constr:
        """ Creates a new constraint (row)

        Adds a new constraint to the model

        Args:
            lin_expr (LinExpr): linear expression
            name (str): optional constraint name, used when saving model to\
            lp or mps files

        Examples:

        The following code adds the constraint :math:`x_1 + x_2 \leq 1` 
        (x1 and x2 should be created first using 
        :func:`add_var<mip.model.Model.add_var>`)::

            m += x1 + x2 <= 1

        Which is equivalent to::

            m.add_constr( x1 + x2 <= 1 )

        Summation expressions can be used also, to add the constraint \
        :math:`\displaystyle \sum_{i=0}^{n-1} x_i = y` and name this \
        constraint cons1::

            m += xsum(x[i] for i in range(n)) == y, "cons1"

        """

        if isinstance(lin_expr, bool):
            return None  # empty constraint
        idx = self.solver.add_constr(lin_expr, name)
        self.constrs.append(Constr(self, idx, name))
        self.constrs_by_name[name] = self.constrs[-1]
        return self.constrs[-1]

    def copy(self, solver_name: str = None) -> "Model":
        """ Creates a copy of the current model

        Args:
            solver_name(str): solver name (optional)

        Returns:
            Model: clone of current model

        """
        if not solver_name:
            solver_name = self.solver_name
        copy = Model(self.name, self.sense, solver_name)

        # adding variables
        for v in self.vars:
            copy.add_var(name=v.name, lb=v.lb, ub=v.ub, obj=v.obj, var_type=v.var_type)

        # adding constraints
        for c in self.constrs:
            expr = c.expr  # todo: make copy of constraint"s lin_expr
            copy.add_constr(lin_expr=expr, name=c.name)

        # setting objective function"s constant
        copy.objective_const = self.get_objective_const()

        return copy

    def get_constr_by_name(self, name: str) -> "Constr":
        """ Queries a constraint per name

        Args:
            name(str): constraint name

        Returns:
            Constr: constraint
        """
        return self.constrs_by_name.get(name, None)

    @property
    def objective_bound(self) -> float:
        return self.solver.get_objective_bound()

    @property
    def objective(self) -> LinExpr:
        """LinExpr: Objective function of the problem

        The objective function of the problem as a linear expression.

        Examples:

            The following code adds all x variables x[0], ..., x[n-1], to
            the objective function of model m with weight w::

                m.objective = xsum(w*x[i] for i in range(n))

            A simpler way to define the objective function is the use of the
            model operator += ::

                m += xsum(w*x[i] for i in range(n))

            Note that the only difference of adding a constraint is the lack of
            a sense and a rhs.

        """
        return self.solver.get_objective()

    @objective.setter
    def objective(self, expr):
        if isinstance(expr, int) or isinstance(expr, float):
            self.solver.set_objective(LinExpr([], [], expr))
        elif isinstance(expr, Var):
            self.solver.set_objective(LinExpr([expr], [1]))
        elif isinstance(expr, LinExpr):
            self.solver.set_objective(expr)

    @property
    def sense(self) -> str:
        """ The optimization sense

        Returns:
            str: the objective function sense, MINIMIZE (default) or (MAXIMIZE)
        """

        return self.solver.get_objective_sense()

    @sense.setter
    def sense(self, sense: str):
        self.solver.set_objective_sense(sense)

    @property
    def objective_const(self) -> float:
        """ Returns the current constant part of the objective function

        float: the constant part in the objective function
        """
        return self.solver.get_objective_const()

    @objective_const.setter
    def objective_const(self, const: float) -> None:
        self.solver.set_objective_const(const)

    @property
    def objective_value(self) -> float:
        """ Objective function value

        Returns:
            float: returns the objective function value of the solution found.

        """
        return self.solver.get_objective_value()

    @property
    def num_solutions(self) -> int:
        """ Number of solutions found during the MIP search

        Returns:
            int: number of solutions stored in the solution pool

        """
        return self.solver.get_num_solutions()

    @property
    def objective_values(self) -> List[float]:
        """ List of costs of all solutions in the solution pool

        Returns:
            List[float]: costs of all solutions stored in the solution pool 
            as an array from 0 (the best solution) to num_solutions-1.
        """
        return [float(self.solver.get_objective_value_i(i))\
                 for i in range(self.num_solutions)] 

    def get_var_by_name(self, name) -> "Var":
        """ Searchers a variable by its name

        Returns:
            Var: a reference to a variable
        """
        return self.vars_by_name.get(name, None)

    def relax(self):
        """ Relax integrality constraints of variables

        Changes the type of all integer and binary variables to
        continuous. Bounds are preserved.
        """
        self.solver.relax()
        for v in self.vars:
            if v.var_type == BINARY or v.var_type == INTEGER:
                v.var_type = CONTINUOUS

    def add_cut_generator(self, cuts_generator: "CutsGenerator") -> None:
        """ Adds a cut generator

        Cut generators are called whenever a solution where one or more
        integer variables appear with continuous values. A cut generator will
        try to produce one or more inequalities to remove this fractional point.

        Args:
            cuts_generator : CutsGenerator

        """
        self.cut_generators.append(cuts_generator)

    @property
    def emphasis(self) -> int:
        """int: defines the main objective of the search, if set to 1 (FEASIBILITY) then
        the search process will focus on try to find quickly feasible solutions and
        improving them; if set to 2 (OPTIMALITY) then the search process will try to
        find a provable optimal solution, procedures to further improve the lower bounds will
        be activated in this setting, this may increase the time to produce the first
        feasible solutions but will probably pay off in longer runs; the default option
        if 0, where a balance between optimality and feasibility is sought.
        """
        return self.solver.get_emphasis()

    @emphasis.setter
    def emphasis(self, emph: int):
        self.solver.set_emphasis(emph)

    def optimize(self,
                 branch_selector: "BranchSelector" = None,
                 incumbent_updater: "IncumbentUpdater" = None,
                 lazy_constrs_generator: "LazyConstrsGenerator" = None,
                 max_seconds: float = inf,
                 max_nodes: int = inf,
                 max_solutions: int = inf) -> int:
        """ Optimizes current model

        Optimizes current model, optionally specifying processing limits.

        To optimize model m within a processing time limit of 300 seconds::

            m.optimize(max_seconds=300)

        Args:
            branch_selector (BranchSelector): Callback to select branch (an object of a class inheriting from BranchSelector must be passed)
            cuts_generator (CutsGenerator): Callback to generate cuts (an object of a class inheriting from CutsGenerator must be passed)
            incumbent_updater (IncumbentUpdater): Callback to update incumbent solution (an object of a class inheriting from IncumbentUpdater must be passed)
            lazy_constrs_generator (LazyConstrsGenerator): Callback to include lazy generated constraints (an object of a class inheriting from LazyConstrsGenerator must be passed)
            max_seconds (float): Maximum runtime in seconds (default: inf)
            max_nodes (float): Maximum number of nodes (default: inf)
            max_solutions (float): Maximum number of solutions (default: inf)

        Returns:
            int: optimization status, which can be OPTIMAL(0), ERROR(-1), INFEASIBLE(1), UNBOUNDED(2). When optimizing problems
            with integer variables some additional cases may happen, FEASIBLE(3) for the case when a feasible solution was found
            but optimality was not proved, INT_INFEASIBLE(4) for the case when the lp relaxation is feasible but no feasible integer
            solution exists and NO_SOLUTION_FOUND(5) for the case when an integer solution was not found in the optimization.

        """
        self.solver.set_callbacks(branch_selector, incumbent_updater, lazy_constrs_generator)
        self.solver.set_processing_limits(max_seconds, max_nodes, max_solutions)

        return self.solver.optimize()

    def read(self, path: str) -> None:
        """ Reads a MIP model

        Reads a MIP model in .lp or .mps file format.

        Args:
            path(str): file name

        """
        self.solver.read(path)
        n_cols = self.solver.num_cols()
        n_rows = self.solver.num_rows()
        for i in range(n_cols):
            self.vars.append(Var(self, i, self.solver.var_get_name(i)))
            self.vars_by_name[self.vars[-1].name] = self.vars[-1]
        for i in range(n_rows):
            self.constrs.append(Constr(self, i, self.solver.constr_get_name(i)))
            self.constrs_by_name[self.constrs[-1].name] = self.constrs[-1]
        self.sense = self.solver.get_objective_sense()

    @property
    def start(self) -> List[Tuple["Var", float]]:
        """ Enter an initial feasible solution

        Enters an initial feasible solution. Only the main binary/integer decision variables.
        Auxiliary or continuous variables are automatically computed.

        Args:
            start_sol: list of tuples Var,float indicating non-zero variables
            and their values in the initial feasible solution

        """
        return self.__mipStart

    @start.setter
    def start(self, start_sol: List[Tuple["Var", float]]):
        self.__mipStart = start_sol
        self.solver.set_start(start_sol)

    def write(self, path: str) -> None:
        """ Saves the the MIP model

        Args:
            path(str): file name

        Saves the the MIP model, use the extension ".lp" or ".mps" in the file
        name to specify the file format.

        """
        self.solver.write(path)

    @property
    def num_cols(self) -> int:
        """int: number of columns (variables) in the model"""
        return len(self.vars)

    @property
    def num_rows(self) -> int:
        """int: number of rows (constraints) in the model"""
        return len(self.constrs)

    @property 
    def num_nz(self) -> int:
        """int: number of non-zeros in the constraint matrix"""
        return self.solver.num_nz()

    @property
    def cutoff(self) -> float:
        """float: upper limit for the solution cost, solutions with cost > cutoff
        will be removed from the search space, a small cutoff value may significantly
        speedup the search, but if cutoff is set to a value too low
        the model will become infeasible"""
        return self.solver.get_cutoff()

    @cutoff.setter
    def cutoff(self, value: float):
        self.solver.set_cutoff(value)

    @property
    def mip_gap_abs(self) -> float:
        """float: tolerance for the quality of the optimal solution, if a
        solution with cost c and a lower bound l are available and c-l<mip_gap_abs,
        the search will be concluded, see mip_gap to determine
        a percentage value """
        return self.solver.get_mip_gap_abs()

    @mip_gap_abs.setter
    def mip_gap_abs(self, value):
        self.solver.set_mip_gap(value)

    @property
    def mip_gap(self) -> float:
        """float: percentage indicating the tolerance for the maximum percentage deviation
        from the optimal solution cost, if a solution with cost c and a lower bound l
        are available and (c-l)/l < mip_gap the search will be concluded."""
        return self.solver.get_mip_gap()

    @mip_gap.setter
    def mip_gap(self, value):
        self.solver.set_mip_gap(value)

    @property
    def max_seconds(self) -> float:
        """float: time limit in seconds for search"""
        return self.solver.get_max_seconds()

    @max_seconds.setter
    def max_seconds(self, max_seconds: float):
        self.solver.set_max_seconds(max_seconds)

    @property
    def max_nodes(self) -> int:
        """int: maximum number of nodes to be explored in the search tree"""
        return self.solver.get_max_nodes()

    @max_nodes.setter
    def max_nodes(self, max_nodes: int):
        self.solver.set_max_nodes(max_nodes)

    @property
    def max_solutions(self) -> int:
        """int: solution limit, search will be stopped when max_solutions were found"""
        return self.solver.get_max_solutions()

    @max_solutions.setter
    def max_solutions(self, max_solutions: int):
        self.solver.set_max_solutions(max_solutions)


class Solver:

    def __init__(self, model: Model, name: str, sense: str):
        self.model = model
        self.name = name
        self.sense = sense

    def __del__(self): pass

    def add_var(self,
                name: str = "",
                obj: float = 0,
                lb: float = 0,
                ub: float = INF,
                var_type: str = CONTINUOUS,
                column: "Column" = None) -> int:
        if var_type == BINARY:
            lb = 0.0
            ub = 1.0

    def add_constr(self, lin_expr: "LinExpr", name: str = "") -> int: pass

    def get_objective_bound(self) -> float: pass

    def get_objective(self) -> LinExpr: pass

    def get_objective_const(self) -> float: pass

    def relax(self): pass

    def optimize(self) -> int: pass

    def get_objective_value(self) -> float: pass

    def get_objective_value_i(self, i: int) -> float: pass

    def get_num_solutions(self) -> int: pass

    def get_objective_sense(self) -> str: pass

    def set_objective_sense(self, sense: str): pass

    def set_start(self, start: List[Tuple["Var", float]]) -> None: pass

    def set_objective(self, lin_expr: "LinExpr", sense: str = "") -> None: pass

    def set_objective_const(self, const: float) -> None: pass

    def set_callbacks(self,
                      branch_selector: "BranchSelector" = None,
                      incumbent_updater: "IncumbentUpdater" = None,
                      lazy_constrs_generator: "LazyConstrsGenerator" = None) -> None:
        pass

    def set_processing_limits(self,
                              max_time: float = inf,
                              max_nodes: int = inf,
                              max_sol: int = inf):
        pass

    def get_max_seconds(self) -> float: pass

    def set_max_seconds(self, max_seconds: float): pass

    def get_max_solutions(self) -> int: pass

    def set_max_solutions(self, max_solutions: int): pass

    def get_max_nodes(self) -> int: pass

    def set_max_nodes(self, max_nodes: int): pass

    def write(self, file_path: str) -> None: pass

    def read(self, file_path: str) -> None: pass

    def num_cols(self) -> int: pass

    def num_rows(self) -> int: pass

    def num_nz(self) -> int:pass

    def get_emphasis(self) -> int: pass

    def set_emphasis(self, emph: int): pass

    def get_cutoff(self) -> float: pass

    def set_cutoff(self, cutoff: float): pass

    def get_mip_gap_abs(self) -> float: pass

    def set_mip_gap_abs(self, mip_gap_abs: float): pass

    def get_mip_gap(self) -> float: pass

    def set_mip_gap(self, mip_gap: float): pass

    # Constraint-related getters/setters

    def constr_get_expr(self, constr: Constr) -> LinExpr: pass

    def constr_set_expr(self, constr: Constr, value: LinExpr) -> LinExpr: pass

    def constr_get_name(self, idx: int) -> str: pass

    def constr_get_pi(self, constr: Constr) -> float: pass

    # Variable-related getters/setters

    def var_get_lb(self, var: "Var") -> float: pass

    def var_set_lb(self, var: "Var", value: float) -> None: pass

    def var_get_ub(self, var: "Var") -> float: pass

    def var_set_ub(self, var: "Var", value: float) -> None: pass

    def var_get_obj(self, var: "Var") -> float: pass

    def var_set_obj(self, var: "Var", value: float) -> None: pass

    def var_get_type(self, var: "Var") -> str: pass

    def var_set_type(self, var: "Var", value: str) -> None: pass

    def var_get_column(self, var: "Var") -> Column: pass

    def var_set_column(self, var: "Var", value: Column) -> None: pass

    def var_get_rc(self, var: "Var") -> float: pass

    def var_get_x(self, var: "Var") -> float: pass

    def var_get_xi(self, var: "Var", i: int) -> float: pass

    def var_get_name(self, idx: int) -> str: pass


class Var:

    def __init__(self,
                 model: Model,
                 idx: int,
                 name: str = ""):
        self.model = model
        self.idx = idx
        self.name = name  # discuss this var

    def __hash__(self) -> int:
        return self.idx

    def __add__(self, other) -> LinExpr:
        if isinstance(other, Var):
            return LinExpr([self, other], [1, 1])
        elif isinstance(other, LinExpr):
            return other.__add__(self)
        elif isinstance(other, int) or isinstance(other, float):
            return LinExpr([self], [1], other)

    def __radd__(self, other) -> LinExpr:
        return self.__add__(other)

    def __sub__(self, other) -> LinExpr:
        if isinstance(other, Var):
            return LinExpr([self, other], [1, -1])
        elif isinstance(other, LinExpr):
            return (-other).__iadd__(self)
        elif isinstance(other, int) or isinstance(other, float):
            return LinExpr([self], [1], -other)

    def __rsub__(self, other) -> LinExpr:
        if isinstance(other, Var):
            return LinExpr([self, other], [-1, 1])
        elif isinstance(other, LinExpr):
            return other.__sub__(self)
        elif isinstance(other, int) or isinstance(other, float):
            return LinExpr([self], [-1], other)

    def __mul__(self, other) -> LinExpr:
        assert isinstance(other, int) or isinstance(other, float)
        return LinExpr([self], [other])

    def __rmul__(self, other) -> LinExpr:
        return self.__mul__(other)

    def __truediv__(self, other) -> LinExpr:
        assert isinstance(other, int) or isinstance(other, float)
        return self.__mul__(1.0 / other)

    def __neg__(self) -> LinExpr:
        return LinExpr([self], [-1.0])

    def __eq__(self, other) -> LinExpr:
        if isinstance(other, Var):
            return LinExpr([self, other], [1, -1], sense="=")
        elif isinstance(other, LinExpr):
            return other == self
        elif isinstance(other, int) or isinstance(other, float):
            if other != 0:
                return LinExpr([self], [1], -1 * other, sense="=")
            return LinExpr([self], [1], sense="=")

    def __le__(self, other) -> LinExpr:
        if isinstance(other, Var):
            return LinExpr([self, other], [1, -1], sense="<")
        elif isinstance(other, LinExpr):
            return other >= self
        elif isinstance(other, int) or isinstance(other, float):
            if other != 0:
                return LinExpr([self], [1], -1 * other, sense="<")
            return LinExpr([self], [1], sense="<")

    def __ge__(self, other) -> LinExpr:
        if isinstance(other, Var):
            return LinExpr([self, other], [1, -1], sense=">")
        elif isinstance(other, LinExpr):
            return other <= self
        elif isinstance(other, int) or isinstance(other, float):
            if other != 0:
                return LinExpr([self], [1], -1 * other, sense=">")
            return LinExpr([self], [1], sense=">")

    def __str__(self) -> str:
        return self.name

    @property
    def lb(self) -> float:
        return self.model.solver.var_get_lb(self)

    @lb.setter
    def lb(self, value: float) -> None:
        self.model.solver.var_set_lb(self, value)

    @property
    def ub(self) -> float:
        return self.model.solver.var_get_ub(self)

    @ub.setter
    def ub(self, value: float) -> None:
        self.model.solver.var_set_ub(self, value)

    @property
    def obj(self) -> float:
        return self.model.solver.var_get_obj(self)

    @obj.setter
    def obj(self, value: float) -> None:
        self.model.solver.var_set_obj(self, value)

    @property
    def type(self) -> str:
        return self.model.solver.var_get_type(self)

    @type.setter
    def type(self, value: str) -> None:
        assert value in (BINARY, CONTINUOUS, INTEGER)
        self.model.solver.var_set_type(self, value)

    @property
    def column(self) -> Column:
        return self.model.solver.var_get_column(self)

    @column.setter
    def column(self, value: Column) -> None:
        self.model.solver.var_set_column(self, value)

    @property
    def rc(self) -> float:
        return self.model.solver.var_get_rc(self)

    @property
    def x(self) -> float:
        return self.model.solver.var_get_x(self)

    def xi(self, i: int) -> float:
        return self.model.solver.var_get_xi(self, i)


class BranchSelector:
    def __init__(self, model: Model):
        self.model = model

    def select_branch(self, relax_solution: List[Tuple[Var, float]]) -> Tuple[Var, int]:
        raise NotImplementedError()


class CutsGenerator:
    def __init__(self, model: Model):
        self.model = model

    def generate_cuts(self, relax_solution: List[Tuple[Var, float]]) -> List[LinExpr]:
        raise NotImplementedError()


class IncumbentUpdater:
    def __init__(self, model: Model):
        self.model = model

    def update_incumbent(self, solution: List[Tuple[Var, float]]) -> List[Tuple[Var, float]]:
        raise NotImplementedError()


class LazyConstrsGenerator:
    def __init(self, model: Model):
        self.model = model

    def generate_lazy_constrs(self, solution: List[Tuple[Var, float]]) -> List[LinExpr]:
        raise NotImplementedError()


def xsum(terms) -> LinExpr:
    result = LinExpr()
    for term in terms:
        result.add_term(term)
    return result


# function aliases
quicksum = xsum


def read_custom_settings() -> str:
    global customCbcLib
    from pathlib import Path
    home = str(Path.home())
    import os
    configpath = os.path.join(home, ".config")
    if os.path.isdir(configpath):
        conffile = os.path.join(configpath, "python-mip")
        if os.path.isfile(conffile):
            f = open(conffile, "r")
            for line in f:
                if "=" in line:
                    cols = line.split("=")
                    if cols[0].strip().lower() == "cbc-library":
                        customCbcLib = cols[1].lstrip().rstrip().replace('"', "")


print("using python mip package version {}".format(VERSION))
customCbcLib = ""
read_custom_settings()
# print("customCbcLib {}".format(customCbcLib))

# vim: ts=4 sw=4 et



VERSION = '1.0.26'

# epsilon number (practical zero)
EPS = 10e-6

# infinity representation
INF = float("inf")

# optimization status
ERROR = -1
OPTIMAL = 0
INFEASIBLE = 1
UNBOUNDED = 2
FEASIBLE = 3
INT_INFEASIBLE = 4
NO_SOLUTION_FOUND = 5
LOADED = 6
CUTOFF = 7

# constraint senses
EQUAL = "="
LESS_OR_EQUAL = "<"
GREATER_OR_EQUAL = ">"

# optimization directions
MINIMIZE = "MIN"
MAXIMIZE = "MAX"

# Search emphasis
FEASIBILITY = 1
OPTIMALITY = 2

# variable types
BINARY = "B"
CONTINUOUS = "C"
INTEGER = "I"

# solvers
CBC = "CBC"
CPLEX = "CPX"
GUROBI = "GRB"
SCIP = "SCIP"



import mip
from mip.model import *
from ctypes import *
from ctypes.util import *
from typing import Dict
from sys import platform
from os.path import dirname

warningMessages = 0


class SolverCbc(Solver):
    def __init__(self, model: Model, name: str, sense: str):
        super().__init__(model, name, sense)

        self._model = cbcNewModel()

        self._objconst = 0.0

        # to not add cut generators twice when reoptimizing
        self.added_cut_callback = False

        # setting objective sense
        if sense == MAXIMIZE:
            cbcSetObjSense(self._model, c_double(-1.0))

        self.emphasis = 0

    def add_var(self,
                obj: float = 0,
                lb: float = 0,
                ub: float = float("inf"),
                coltype: str = "C",
                column: "Column" = None,
                name: str = "") -> int:
        # collecting column data
        numnz = 0 if column is None else len(column.constrs)
        vind = (c_int * numnz)()
        vval = (c_double * numnz)()

        # collecting column coefficients
        for i in range(numnz):
            vind[i] = column.constrs[i].idx
            vval[i] = column.coeffs[i]

        isInt = \
            c_char(1) if coltype.upper() == "B" or coltype.upper() == "I"\
            else c_char(0)

        idx = int(cbcNumCols(self._model))

        cbcAddCol(self._model, c_str(name),
                  c_double(lb), c_double(ub), c_double(obj),
                  isInt, numnz, vind, vval)

        return idx

    def get_objective_const(self) -> float:
        return self._objconst

    def set_objective(self, lin_expr: "LinExpr", sense: str = "") -> None:
        # collecting variable coefficients
        for var, coeff in lin_expr.expr.items():
            cbcSetObjCoeff(self._model, c_int(var.idx), c_double(coeff))

        # objective function constant
        self._objconst = c_double(lin_expr.const)

        # setting objective sense
        if sense == MAXIMIZE:
            cbcSetObjSense(self._model, c_double(-1.0))
        elif sense == MINIMIZE:
            cbcSetObjSense(self._model, c_double(1.0))

    def relax(self):
        for var in self.model.vars:
            if cbcIsInteger(self._model, var.idx):
                cbcSetContinuous(self._model, c_int(var.idx))

    def get_max_seconds(self) -> float:
        return cbcGetMaxSeconds(self._model)

    def set_max_seconds(self, max_seconds: float):
        cbcSetMaxSeconds(self._model, c_double(max_seconds))

    def get_max_solutions(self) -> int:
        return cbcGetMaxSolutions(self._model)

    def set_max_solutions(self, max_solutions: int):
        cbcSetMaxSolutions(self._model, c_int(max_solutions))

    def get_max_nodes(self) -> int:
        return cbcGetMaxNodes(self._model)

    def set_max_nodes(self, max_nodes: int):
        cbcSetMaxNodes(self._model, c_int(max_nodes))

    def optimize(self) -> int:
        # get name indexes from an osi problem
        def cbc_get_osi_name_indexes(osiSolver: c_void_p) -> Dict[str, int]:
            nameIdx = {}
            n = osiNumCols(osiSolver)
            nameSpace = create_string_buffer(256)
            for i in range(n):
                osiColName(osiSolver, c_int(i), nameSpace, 255)
                cname = nameSpace.value.decode('utf-8')
                nameIdx[cname] = i

            return nameIdx

        # cut callback
        def cbc_cut_callback(osiSolver: c_void_p, osiCuts: c_void_p) -> None:
            global warningMessages
            # getting fractional solution
            fracSol = []
            n = osiNumCols(osiSolver)
            nameSpace = create_string_buffer(256)
            x = osiColSolution(osiSolver)
            if x == c_void_p(0):
                raise Exception('no solution found')

            for i in range(n):
                val = float(x[i])
                if abs(val) < 1e-7:
                    continue

                osiColName(osiSolver, c_int(i), nameSpace, 255)
                cname = nameSpace.value.decode('utf-8')
                var = self.model.get_var_by_name(cname)
                if var is None:
                    print('-->> var {} not found'.format(cname))
                fracSol.append((var, val))

            # storing names and indexes to translate cuts
            nameIdx = {}
            cidx = (c_int * n)()
            cval = (c_double * n)()

            # calling cut generators
            for cg in self.model.cut_generators:
                cuts = cg.generate_cuts(fracSol)

                if cuts and not nameIdx:
                    nameIdx = cbc_get_osi_name_indexes(osiSolver)

                # translating cuts for variables in the preprocessed problem
                for cut in cuts:
                    cutIdx = []
                    cutCoef = []
                    hasAllVars = True
                    missingVarName = ''
                    for v, c in cut.expr.items():
                        if v.name in nameIdx.keys():
                            cutIdx.append(nameIdx[v.name])
                            cutCoef.append(c)
                        else:
                            hasAllVars = False
                            missingVarName = v.name
                            break
                    if hasAllVars:
                        nz = len(cutIdx)
                        for i, ci in enumerate(cutIdx):
                            cidx[i] = ci
                            cval[i] = cutCoef[i]
                        sense = c_char(ord(cut.sense))
                        rhs = c_double(-cut.const)
                        osiCutsAddRowCut(osiCuts, c_int(nz), cidx, cval,
                                         sense, rhs)
                        # print('cut add successfully')
                    else:
                        if warningMessages < 5:
                            print('cut discarded because variable {} does not \
                            exists in preprocessed problem.'.format(
                                missingVarName))
                            warningMessages += 1

        # adding cut generators
        if self.model.cut_generators and self.added_cut_callback is False:
            self._cutCallback = CBCcallbacktype(cbc_cut_callback)
            cbcAddCutCallback(self._model, self._cutCallback,
                              c_str("mipCutGen"), c_void_p(0))
            self.added_cut_callback = True

        if self.emphasis == FEASIBILITY:
            cbcSetParameter(self._model, c_str('passf'), c_str('50'))
            cbcSetParameter(self._model, c_str('proximity'), c_str('on'))
        if self.emphasis == OPTIMALITY:
            cbcSetParameter(self._model, c_str('strong'), c_str('10'))
            cbcSetParameter(self._model, c_str('trust'), c_str('20'))
            cbcSetParameter(self._model, c_str('lagomory'), c_str('endonly'))
            cbcSetParameter(self._model, c_str('latwomir'), c_str('endonly'))

        cbcSetParameter(self._model, c_str('maxSavedSolutions'), c_str('10'))
        cbcSolve(self._model)

        if cbcIsAbandoned(self._model):
            return ERROR

        if cbcIsProvenOptimal(self._model):
            return OPTIMAL

        if cbcIsProvenInfeasible(self._model):
            return INFEASIBLE

        if cbcIsContinuousUnbounded(self._model):
            return UNBOUNDED

        if cbcNumIntegers(self._model):
            if cbcBestSolution(self._model):
                return FEASIBLE

        return INFEASIBLE

    def get_objective_sense(self) -> str:
        obj = cbcGetObjSense(self._model)
        if obj < 0.0:
            return MAXIMIZE

        return MINIMIZE

    def set_objective_sense(self, sense: str):
        if sense.strip().upper() == MAXIMIZE.strip().upper():
            cbcSetObjSense(self._model, c_double(-1.0))
        elif sense.strip().upper() == MINIMIZE.strip().upper():
            cbcSetObjSense(self._model, c_double(1.0))
        else:
            raise Exception("Unknown sense: {}, use {} or {}".format(sense,
                                                                     MAXIMIZE,
                                                                     MINIMIZE))

    def get_objective_value(self) -> float:
        return cbcObjValue(self._model)

    def get_objective_bound(self) -> float:
        return cbcGetObjBound(self._model)

    def var_get_x(self, var: Var) -> float:
        if cbcNumIntegers(self._model) > 0:
            x = cbcBestSolution(self._model)
            if x == c_void_p(0):
                raise Exception('no solution found')
            return float(x[var.idx])
        else:
            x = cbcColSolution(self._model)
            return float(x[var.idx])

    def get_num_solutions(self) -> int:
        return cbcNumberSavedSolutions(self._model)

    def get_objective_value_i(self, i: int) -> float:
        return float(cbcSavedSolutionObj(self._model, c_int(i)))

    def var_get_xi(self, var: "Var", i: int) -> float:
        x = cbcSavedSolution(self._model, c_int(i))
        if x == c_void_p(0):
            raise Exception('no solution found')
        return float(x[var.idx])

    def var_get_rc(self, var: Var) -> float:
        rc = cbcReducedCost(self._model)
        return float(rc[var.idx])

    def var_get_lb(self, var: "Var") -> float:
        res = float(cbcGetColLower(self._model)[var.idx])
        return res

    def var_get_ub(self, var: "Var") -> float:
        res = float(cbcGetColUpper(self._model)[var.idx])
        return res

    def var_get_name(self, idx: int) -> str:
        nameSpace = create_string_buffer(256)
        cbcGetColName(self._model, c_int(idx), nameSpace, 255)
        return nameSpace.value.decode('utf-8')

    def var_get_obj(self, var: Var) -> float:
        return float(cbcGetObjCoeff(self._model)[var.idx])

    def var_get_type(self, var: "Var") -> str:
        isInt = cbcIsInteger(self._model, c_int(var.idx))
        if isInt:
            lb = self.var_get_lb(var)
            ub = self.var_get_ub(var)
            if abs(lb) <= 1e-15 and abs(ub - 1.0) <= 1e-15:
                return BINARY
            else:
                return INTEGER

        return CONTINUOUS

    def var_get_column(self, var: "Var") -> Column:
        numnz = cbcGetColNz(self._model, var.idx)

        cidx = cbcGetColIndices(self._model, var.idx)
        ccoef = cbcGetColCoeffs(self._model, var.idx)

        col = Column()

        for i in range(numnz):
            col.constrs.append(self.model.constrs[cidx[i]])
            col.coeffs.append(ccoef[i])

        return col

    def add_constr(self, lin_expr: "LinExpr", name: str = "") -> int:
        # collecting linear expression data
        numnz = len(lin_expr.expr)
        cind = (c_int * numnz)()
        cval = (c_double * numnz)()

        # collecting variable coefficients
        for i, (var, coeff) in enumerate(lin_expr.expr.items()):
            cind[i] = var.idx
            cval[i] = coeff

        # constraint sense and rhs
        sense = c_char(ord(lin_expr.sense))
        rhs = c_double(-lin_expr.const)

        # constraint index
        idx = int(cbcNumRows(self._model))

        cbcAddRow(self._model, c_str(name), numnz, cind, cval, sense, rhs)

        return idx

    def write(self, file_path: str):
        if ".mps" in file_path.lower():
            cbcWriteMps(self._model, c_str(file_path))
        else:
            cbcWriteLp(self._model, c_str(file_path))

    def read(self, file_path: str) -> None:
        if ".mps" in file_path.lower():
            cbcReadMps(self._model, c_str(file_path))
        else:
            cbcReadLp(self._model, c_str(file_path))

    def set_start(self, start: List[Tuple["Var", float]]) -> None:
        n = len(start)
        count = c_int(n)
        dvalues = (c_double * n)()
        for i in range(n):
            dvalues[i] = start[i][1]
        cidxs = (c_int * n)()
        for i in range(n):
            cidxs[i] = start[i][0].idx

        cbcSetMIPStartI(self._model, count, cidxs, dvalues)

    def num_cols(self) -> int:
        return cbcNumCols(self._model)

    def num_rows(self) -> int:
        return cbcNumRows(self._model)

    def num_nz(self) -> int:
        return cbcNumElements(self._model)

    def get_cutoff(self) -> float:
        return cbcGetCutoff(self._model)

    def set_cutoff(self, cutoff: float):
        cbcSetCutoff(self._model, c_double(cutoff))

    def get_mip_gap_abs(self) -> float:
        return cbcGetAllowableGap(self._model)

    def set_mip_gap_abs(self, allowable_gap: float):
        cbcSetAllowableGap(self._model, c_double(allowable_gap))

    def get_mip_gap(self) -> float:
        return cbcGetAllowableFractionGap(self._model)

    def set_mip_gap(self, allowable_ratio_gap: float):
        cbcSetAllowableFractionGap(self._model, c_double(allowable_ratio_gap))

    def constr_get_expr(self, constr: Constr) -> LinExpr:
        numnz = cbcGetRowNz(self._model, constr.idx)

        ridx = cbcGetRowIndices(self._model, constr.idx)
        rcoef = cbcGetRowCoeffs(self._model, constr.idx)

        rhs = cbcGetRowRHS(self._model, constr.idx)
        rsense = cbcGetRowSense(self._model, constr.idx).\
            decode('utf-8').upper()

        sense = ''
        if (rsense == 'E'):
            sense = EQUAL
        elif (rsense == 'L'):
            sense = LESS_OR_EQUAL
        elif (rsense == 'G'):
            sense = GREATER_OR_EQUAL
        else:
            raise Exception('Unknow sense: {}'.format(rsense))

        expr = LinExpr(const=-rhs, sense=sense)
        for i in range(numnz):
            expr.add_var(self.model.vars[ridx[i]], rcoef[i])

        return expr

    def constr_get_name(self, idx: int) -> str:
        nameSpace = create_string_buffer(256)
        cbcGetRowName(self._model, c_int(idx), nameSpace, 255)
        return nameSpace.value.decode('utf-8')

    def set_processing_limits(self,
                              maxTime=inf,
                              maxNodes=inf,
                              maxSol=inf):
        m = self._model
        if maxTime != inf:
            cbcSetParameter(m, c_str('timeMode'), c_str('elapsed'))
            cbcSetParameter(m, c_str('seconds'), c_str('{}'.format(maxTime)))
        if maxNodes != inf:
            cbcSetParameter(m, c_str('maxNodes'), c_str('{}'.format(maxNodes)))
        if maxSol != inf:
            cbcSetParameter(m, c_str('maxSolutions'),
                            c_str('{}'.format(maxSol)))

    def get_emphasis(self) -> int:
        return self.emphasis

    def set_emphasis(self, emph: int):
        self.emphasis = emph

    def __del__(self):
        cbcDeleteModel(self._model)


has_cbc = False


try:
    if customCbcLib:
        print('CBC library path from config file: {}'.format(
            mip.model.customCbcLib))
        cbclib = CDLL(mip.model.customCbcLib)
        print('has cbc')
        has_cbc = True
    else:
        try:
            pathmip = dirname(mip.__file__)
            pathlib = os.path.join(pathmip, 'libraries')
            if 'linux' in sys.platform.lower():
                if sizeof(c_void_p) == 8:
                    libfile = os.path.join(pathlib, 'cbc-c-linux-x86-64.so')
                    cbclib = CDLL(libfile)
                    has_cbc = True
            elif sys.platform.lower().startswith('win'):
                if sizeof(c_void_p) == 8:
                    libfile = os.path.join(pathlib, 'cbc-c-windows-x86-64.dll')
                    cbclib = CDLL(libfile)
                    has_cbc = True
                else:
                    libfile = os.path.join(pathlib, 'cbc-c-windows-x86-32.dll')
                    cbclib = CDLL(libfile)
                    has_cbc = True
            elif sys.platform.lower().startswith('darwin'):
                if sizeof(c_void_p) == 8:
                    libfile = os.path.join(pathlib, 'cbc-c-darwin-x86-64.a')
                    cbclib = CDLL(libfile)
                    has_cbc = True
        except:
            try:
                # linux library
                cbclib = CDLL(find_library("CbcSolver"))
                has_cbc = True
                print('cbc found')
            except:
                # window library
                try:
                    cbclib = CDLL(find_library("cbcCInterfaceDll"))
                    has_cbc = True
                    print('cbc found')
                except:
                    try:
                        cbclib = CDLL(find_library("./cbcCInterfaceDll"))
                        has_cbc = True
                        print('cbc found')
                    except:
                        has_cbc = False
except:
    has_cbc = False
    print('cbc not found')

if has_cbc:
    method_check = ""
    try:
        method_check = "Cbc_newModel"
        cbcNewModel = cbclib.Cbc_newModel
        cbcNewModel.restype = c_void_p

        method_check = "Cbc_readLp"
        cbcReadLp = cbclib.Cbc_readLp
        cbcReadLp.argtypes = [c_void_p, c_char_p]
        cbcReadLp.restype = c_int

        method_check = "Cbc_readMps"
        cbcReadMps = cbclib.Cbc_readMps
        cbcReadMps.argtypes = [c_void_p, c_char_p]
        cbcReadMps.restype = c_int

        method_check = "Cbc_writeLp"
        cbcWriteLp = cbclib.Cbc_writeLp
        cbcWriteLp.argtypes = [c_void_p, c_char_p]

        method_check = "Cbc_writeMps"
        cbcWriteMps = cbclib.Cbc_writeMps
        cbcWriteMps.argtypes = [c_void_p, c_char_p]

        method_check = "Cbc_getNumCols"
        cbcNumCols = cbclib.Cbc_getNumCols
        cbcNumCols.argtypes = [c_void_p]
        cbcNumCols.restype = c_int

        method_check = "Cbc_getNumIntegers"
        cbcNumIntegers = cbclib.Cbc_getNumIntegers
        cbcNumIntegers.argtypes = [c_void_p]
        cbcNumIntegers.restype = c_int

        method_check = "Cbc_getNumRows"
        cbcNumRows = cbclib.Cbc_getNumRows
        cbcNumRows.argtypes = [c_void_p]
        cbcNumRows.restype = c_int

        method_check = "Cbc_getNumElements"
        cbcNumElements = cbclib.Cbc_getNumElements
        cbcNumElements.argtypes = [c_void_p]
        cbcNumElements.restype = c_int

        method_check = "Cbc_getRowNz"
        cbcGetRowNz = cbclib.Cbc_getRowNz
        cbcGetRowNz.argtypes = [c_void_p, c_int]
        cbcGetRowNz.restype = c_int

        method_check = "Cbc_getRowIndices"
        cbcGetRowIndices = cbclib.Cbc_getRowIndices
        cbcGetRowIndices.argtypes = [c_void_p, c_int]
        cbcGetRowIndices.restype = POINTER(c_int)

        method_check = "Cbc_getRowCoeffs"
        cbcGetRowCoeffs = cbclib.Cbc_getRowCoeffs
        cbcGetRowCoeffs.argtypes = [c_void_p, c_int]
        cbcGetRowCoeffs.restype = POINTER(c_double)

        method_check = "Cbc_getRowRHS"
        cbcGetRowRHS = cbclib.Cbc_getRowRHS
        cbcGetRowRHS.argtypes = [c_void_p, c_int]
        cbcGetRowRHS.restype = c_double

        method_check = "Cbc_getRowSense"
        cbcGetRowSense = cbclib.Cbc_getRowSense
        cbcGetRowSense.argtypes = [c_void_p, c_int]
        cbcGetRowSense.restype = c_char

        method_check = "Cbc_getColNz"
        cbcGetColNz = cbclib.Cbc_getColNz
        cbcGetColNz.argtypes = [c_void_p, c_int]
        cbcGetColNz.restype = c_int

        method_check = "Cbc_getColIndices"
        cbcGetColIndices = cbclib.Cbc_getColIndices
        cbcGetColIndices.argtypes = [c_void_p, c_int]
        cbcGetColIndices.restype = POINTER(c_int)

        method_check = "Cbc_getColCoeffs"
        cbcGetColCoeffs = cbclib.Cbc_getColCoeffs
        cbcGetColCoeffs.argtypes = [c_void_p, c_int]
        cbcGetColCoeffs.restype = POINTER(c_double)

        method_check = "Cbc_addCol"
        cbcAddCol = cbclib.Cbc_addCol
        cbcAddCol.argtypes = [c_void_p, c_char_p, c_double,
                              c_double, c_double, c_char, c_int,
                              POINTER(c_int), POINTER(c_double)]

        method_check = "Cbc_addRow"
        cbcAddRow = cbclib.Cbc_addRow
        cbcAddRow.argtypes = [c_void_p, c_char_p, c_int,
                              POINTER(c_int), POINTER(c_double), c_char,
                              c_double]

        method_check = "Cbc_setObjCoeff"
        cbcSetObjCoeff = cbclib.Cbc_setObjCoeff
        cbcSetObjCoeff.argtypes = [c_void_p, c_int, c_double]

        method_check = "Cbc_getObjSense"
        cbcGetObjSense = cbclib.Cbc_getObjSense
        cbcGetObjSense.argtypes = [c_void_p]
        cbcGetObjSense.restype = c_double

        method_check = "Cbc_getObjCoefficients"
        cbcGetObjCoeff = cbclib.Cbc_getObjCoefficients
        cbcGetObjCoeff.argtypes = [c_void_p]
        cbcGetObjCoeff.restype = POINTER(c_double)

        method_check = "Cbc_deleteModel"
        cbcDeleteModel = cbclib.Cbc_deleteModel
        cbcDeleteModel.argtypes = [c_void_p]

        method_check = "Cbc_solve"
        cbcSolve = cbclib.Cbc_solve
        cbcSolve.argtypes = [c_void_p]
        cbcSolve.restype = c_int

        method_check = "Cbc_getColSolution"
        cbcColSolution = cbclib.Cbc_getColSolution
        cbcColSolution.argtypes = [c_void_p]
        cbcColSolution.restype = POINTER(c_double)

        method_check = "Cbc_getReducedCost"
        cbcReducedCost = cbclib.Cbc_getReducedCost
        cbcReducedCost.argtypes = [c_void_p]
        cbcReducedCost.restype = POINTER(c_double)

        method_check = "Cbc_bestSolution"
        cbcBestSolution = cbclib.Cbc_bestSolution
        cbcBestSolution.argtypes = [c_void_p]
        cbcBestSolution.restype = POINTER(c_double)

        method_check = "Cbc_numberSavedSolutions"
        cbcNumberSavedSolutions = cbclib.Cbc_numberSavedSolutions
        cbcNumberSavedSolutions.argtypes = [c_void_p]
        cbcNumberSavedSolutions.restype = c_int

        method_check = "Cbc_savedSolution"
        cbcSavedSolution = cbclib.Cbc_savedSolution
        cbcSavedSolution.argtypes = [c_void_p, c_int]
        cbcSavedSolution.restype = POINTER(c_double)

        method_check = "Cbc_savedSolutionObj"
        cbcSavedSolutionObj = cbclib.Cbc_savedSolutionObj
        cbcSavedSolutionObj.argtypes = [c_void_p, c_int]
        cbcSavedSolutionObj.restype = c_double

        method_check = "Cbc_getObjValue"
        cbcObjValue = cbclib.Cbc_getObjValue
        cbcObjValue.argtypes = [c_void_p]
        cbcObjValue.restype = c_double

        method_check = "Cbc_setObjSense"
        cbcSetObjSense = cbclib.Cbc_setObjSense
        cbcSetObjSense.argtypes = [c_void_p, c_double]

        method_check = "Cbc_isProvenOptimal"
        cbcIsProvenOptimal = cbclib.Cbc_isProvenOptimal
        cbcIsProvenOptimal.argtypes = [c_void_p]
        cbcIsProvenOptimal.restype = c_int

        method_check = "Cbc_isProvenInfeasible"
        cbcIsProvenInfeasible = cbclib.Cbc_isProvenInfeasible
        cbcIsProvenInfeasible.argtypes = [c_void_p]
        cbcIsProvenInfeasible.restype = c_int

        method_check = "Cbc_isContinuousUnbounded"
        cbcIsContinuousUnbounded = cbclib.Cbc_isContinuousUnbounded
        cbcIsContinuousUnbounded.argtypes = [c_void_p]
        cbcIsContinuousUnbounded.restype = c_int

        method_check = "Cbc_isAbandoned"
        cbcIsAbandoned = cbclib.Cbc_isAbandoned
        cbcIsAbandoned.argtypes = [c_void_p]
        cbcIsAbandoned.restype = c_int

        method_check = "Cbc_getColLower"
        cbcGetColLower = cbclib.Cbc_getColLower
        cbcGetColLower.argtypes = [c_void_p]
        cbcGetColLower.restype = POINTER(c_double)

        method_check = "Cbc_getColUpper"
        cbcGetColUpper = cbclib.Cbc_getColUpper
        cbcGetColUpper.argtypes = [c_void_p]
        cbcGetColUpper.restype = POINTER(c_double)

        method_check = "Cbc_setColLower"
        cbcSetColLower = cbclib.Cbc_setColLower
        cbcSetColLower.argtypes = [c_void_p, c_int, c_double]
        cbcSetColLower.restype = POINTER(c_double)

        method_check = "Cbc_setColUpper"
        cbcSetColUpper = cbclib.Cbc_setColUpper
        cbcSetColUpper.argtypes = [c_void_p, c_int, c_double]
        cbcSetColUpper.restype = POINTER(c_double)

        method_check = "Cbc_getColName"
        cbcGetColName = cbclib.Cbc_getColName
        cbcGetColName.argtypes = [c_void_p, c_int, c_char_p, c_int]

        method_check = "Cbc_getRowName"
        cbcGetRowName = cbclib.Cbc_getRowName
        cbcGetRowName.argtypes = [c_void_p, c_int, c_char_p, c_int]

        method_check = "Cbc_isInteger"
        cbcIsInteger = cbclib.Cbc_isInteger
        cbcIsInteger.argtypes = [c_void_p, c_int]
        cbcIsInteger.restype = c_int

        method_check = "Cbc_setContinuous"
        cbcSetContinuous = cbclib.Cbc_setContinuous
        cbcSetContinuous.argtypes = [c_void_p, c_int]

        method_check = "Cbc_setParameter"
        cbcSetParameter = cbclib.Cbc_setParameter
        cbcSetParameter.argtypes = [c_void_p, c_char_p, c_char_p]

        method_check = "Cbc_setMIPStart"
        cbcSetMIPStart = cbclib.Cbc_setMIPStart
        cbcSetMIPStart.argtypes = [c_int, POINTER(c_char_p), POINTER(c_double)]

        method_check = "Cbc_setMIPStartI"
        cbcSetMIPStartI = cbclib.Cbc_setMIPStartI
        cbcSetMIPStartI.argtypes = [c_void_p, c_int, POINTER(c_int),
                                    POINTER(c_double)]

        method_check = "CBCCutCallback"
        CBCcallbacktype = CFUNCTYPE(c_void_p, c_void_p, c_void_p)

        method_check = "Cbc_addCutCallback"
        cbcAddCutCallback = cbclib.Cbc_addCutCallback
        cbcAddCutCallback.argtypes = [c_void_p, CBCcallbacktype, c_char_p,
                                      c_void_p]

        method_check = "Osi_getNumCols"
        osiNumCols = cbclib.Osi_getNumCols
        osiNumCols.argtypes = [c_void_p]
        osiNumCols.restype = c_int

        method_check = "Osi_getColName"
        osiColName = cbclib.Osi_getColName
        osiColName.argtypes = [c_void_p, c_int, c_char_p, c_int]

        method_check = "Osi_getColSolution"
        osiColSolution = cbclib.Osi_getColSolution
        osiColSolution.argtypes = [c_void_p]
        osiColSolution.restype = POINTER(c_double)

        method_check = "oc_addRowCut"
        osiCutsAddRowCut = cbclib.OsiCuts_addRowCut
        osiCutsAddRowCut.argtypes = [c_void_p, c_int, POINTER(c_int),
                                     POINTER(c_double), c_char, c_double]

        method_check = "Cbc_getCutoff"
        cbcGetCutoff = cbclib.Cbc_getCutoff
        cbcGetCutoff.argtype = [c_void_p]
        cbcGetCutoff.restype = c_double

        method_check = "Cbc_setCutoff"
        cbcSetCutoff = cbclib.Cbc_setCutoff
        cbcSetCutoff.argtype = [c_void_p, c_double]

        method_check = "Cbc_getAllowableGap"
        cbcGetAllowableGap = cbclib.Cbc_getAllowableGap
        cbcGetAllowableGap.argtype = [c_void_p]
        cbcGetAllowableGap.restype = c_double

        method_check = "Cbc_setAllowableGap"
        cbcSetAllowableGap = cbclib.Cbc_setAllowableGap
        cbcSetAllowableGap.argtype = [c_void_p, c_double]

        method_check = "Cbc_getAllowableFractionGap"
        cbcGetAllowableFractionGap = cbclib.Cbc_getAllowableFractionGap
        cbcGetAllowableFractionGap.argtype = [c_void_p]
        cbcGetAllowableFractionGap.restype = c_double

        method_check = "Cbc_setAllowableFractionGap"
        cbcSetAllowableFractionGap = cbclib.Cbc_setAllowableFractionGap
        cbcSetAllowableFractionGap.argtype = [c_void_p, c_double]

        method_check = "Cbc_getMaximumSeconds"
        cbcGetMaxSeconds = cbclib.Cbc_getMaximumSeconds
        cbcGetMaxSeconds.argtype = [c_void_p]
        cbcGetMaxSeconds.restype = c_double

        method_check = "Cbc_getMaximumNodes"
        cbcGetMaxNodes = cbclib.Cbc_getMaximumNodes
        cbcGetMaxNodes.argtype = [c_void_p]
        cbcGetMaxNodes.restype = c_int

        method_check = "Cbc_getMaximumSolutions"
        cbcGetMaxSolutions = cbclib.Cbc_getMaximumSolutions
        cbcGetMaxSolutions.argtype = [c_void_p]
        cbcGetMaxSolutions.restype = c_int

        method_check = "Cbc_setMaximumSeconds"
        cbcSetMaxSeconds = cbclib.Cbc_setMaximumSeconds
        cbcSetMaxSeconds.argtypes = [c_void_p, c_double]

        method_check = "Cbc_setMaximumNodes"
        cbcSetMaxNodes = cbclib.Cbc_setMaximumNodes
        cbcSetMaxNodes.argtypes = [c_void_p, c_int]

        method_check = "Cbc_setMaximumSolutions"
        cbcSetMaxSolutions = cbclib.Cbc_setMaximumSolutions
        cbcSetMaxSolutions.argtypes = [c_void_p, c_int]

        method_check = "Cbc_getBestPossibleObjValue"
        cbcGetObjBound = cbclib.Cbc_getBestPossibleObjValue
        cbcGetObjBound.argtypes = [c_void_p]
        cbcGetObjBound.restype = c_double

        has_cbc = True
    except:
        print('\nplease install a more updated version of cbc (or cbc trunk),\
 function {} not implemented in the installed version'.format(method_check))
        has_cbc = False


def c_str(value) -> c_char_p:
    """
    This function converts a python string into a C compatible char[]
    :param value: input string
    :return: string converted to C"s format
    """
    return create_string_buffer(value.encode("utf-8"))

# vim: ts=4 sw=4 et



import setuptools


# read the version number 
version_dict = {}
exec(open('mip/constants.py').read(), version_dict)
VERSION = version_dict['VERSION']

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mip",
    version=VERSION,
    author="Toffolo, T.A.M. and Santos, H.G.",
    author_email="haroldo@ufop.edu.br",
    description="Python tools for Modeling and Solving Mixed-Integer Programs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords = ["Optimization", "Linear Programming", "Integer Programming", "Operations Research"],
    url="https://github.com/coin-or/python-mip",
    packages=['mip', 'mip.libraries'],
	package_data = {'mip.libraries' : ['*', '*.*']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from mip.model import *
from ctypes import *
from ctypes.util import *
from math import inf


class SolverGurobi(Solver):

    def __init__(self, model: Model, name: str, sense: str):
        super().__init__(model, name, sense)

        # setting class members to default values
        self._updated = False
        self._num_vars = 0
        self._num_constrs = 0
        self._log = ""
        self._env = c_void_p(0)
        self._model = c_void_p(0)
        self._callback = None

        # setting variables for an empty model
        numvars = c_int(0)
        obj = c_double(0)
        lb = c_double(0)
        ub = c_double(0)
        vtype = c_char_p()
        varnames = c_void_p(0)

        # creating Gurobi environment
        if GRBloadenv(byref(self._env), c_str(self._log)) != 0:
            # todo: raise exception when environment can't be loaded
            pass

        # creating Gurobi model
        if GRBnewmodel(self._env, byref(self._model), c_str(name), numvars,
                       byref(obj), byref(lb), byref(ub), vtype, varnames) != 0:
            # todo: raise exception when environment can't be generated
            pass

        # setting objective sense
        if sense == MAXIMIZE:
            GRBsetintattr(self._model, c_str("ModelSense"), -1)
        elif sense == MINIMIZE:
            GRBsetintattr(self._model, c_str("ModelSense"), 1)

    def __del__(self):
        # freeing Gurobi model and environment
        if self._model:
            GRBfreemodel(self._model)
        if self._env:
            GRBfreeenv(self._env)

    def add_var(self,
                obj: float = 0,
                lb: float = 0,
                ub: float = float("inf"),
                type: str = "C",
                column: "Column" = None,
                name: str = "") -> int:
        # collecting column data
        numnz = 0 if column is None else len(column.constrs)
        vind = (c_int * numnz)()
        vval = (c_double * numnz)()

        # collecting column coefficients
        for i in range(numnz):
            vind[i] = column.constrs[i].idx
            vval[i] = column.coeffs[i]

        # variable type
        vtype = c_char(ord(type))

        # variable index
        idx = self._num_vars
        self._num_vars += 1

        GRBaddvar(self._model, c_int(numnz), vind, vval, c_double(obj), c_double(lb), c_double(ub),
                  vtype, c_str(name))
        self._updated = False

        return idx

    def add_constr(self, lin_expr: LinExpr, name: str = "") -> int:
        # collecting linear expression data
        numnz = len(lin_expr.expr)
        cind = (c_int * numnz)()
        cval = (c_double * numnz)()

        # collecting variable coefficients
        for i, (var, coeff) in enumerate(lin_expr.expr.items()):
            cind[i] = var.idx
            cval[i] = coeff

        # constraint sense and rhs
        sense = c_char(ord(lin_expr.sense))
        rhs = c_double(-lin_expr.const)

        # constraint index
        idx = self._num_constrs
        self._num_constrs += 1

        GRBaddconstr(self._model, numnz, cind, cval, sense, rhs, c_str(name))
        self._updated = False

        return idx

    def get_objective_bound(self) -> float:
        res = c_double()
        GRBgetdblattr(self._model, c_str("ObjBound"), byref(res))
        return res.value

    def get_objective(self) -> LinExpr:
        res = c_double()
        GRBgetdblattr(self._model, c_str("ObjVal"), byref(res))
        return res.value

    def get_objective_const(self) -> float:
        res = c_double()
        GRBgetdblattr(self._model, c_str("ObjCon"), byref(res))
        return res.value

    def relax(self):
        idxs = list()
        for var in self.model.vars:
            vtype = self.var_get_type(var)
            if vtype == BINARY or vtype == INTEGER:
                idxs.append(var.idx)

        ccont = (c_char * len(idxs))()
        for i in range(len(idxs)):
            ccont[i] = CONTINUOUS.encode("utf-8")

        GRBsetcharattrarray(self._model, c_str("VType"), 0, len(idxs), ccont)
        GRBupdatemodel(self._model)

    def get_max_seconds(self) -> float:
        res = c_double()
        GRBgetdblparam(GRBgetenv(self._model), c_str("TimeLimit"), byref(res))
        return res.value

    def set_max_seconds(self, max_seconds: float):
        st = GRBsetdblparam(GRBgetenv(self._model), c_str("TimeLimit"), c_double(max_seconds))
        assert st == 0

    def get_max_solutions(self) -> int:
        res = c_double()
        st = GRBgetintparam(GRBgetenv(self._model), c_str("SolutionLimit"), byref(res))
        assert st == 0
        return res.value

    def set_max_solutions(self, max_solutions: int):
        st = GRBsetintparam(GRBgetenv(self._model), c_str("SolutionLimit"), c_int(max_solutions))
        assert st == 0

    def get_max_nodes(self) -> int:
        res = c_double()
        st = GRBgetdblparam(GRBgetenv(self._model), c_str("NodeLimit"), byref(res))
        assert st == 0
        rdbl = res.value
        rint = min(sys.maxsize, int(rdbl))
        return rint

    def set_max_nodes(self, max_nodes: int):
        st = GRBsetdblparam(GRBgetenv(self._model), c_str("NodeLimit"), c_double(max_nodes))
        assert st == 0

    def optimize(self) -> int:
        # todo add branch_selector and incumbent_updater callbacks
        def callback(p_model: c_void_p,
                     p_cbdata: c_void_p,
                     where: int,
                     p_usrdata: c_void_p) -> int:
            if self.model.cut_generators and where == 5:  # MIPNODE == 5
                # obtaining relaxation solution and "translating" it
                solution = (c_double * self._num_vars)()
                GRBcbget(p_cbdata, where, GRB_CB_MIPNODE_REL, solution)
                relax_solution = []
                for i in range(self._num_vars):
                    if solution[i] <= -EPS or solution[i] >= EPS:
                        relax_solution.append((self.model.vars[i], solution[i]))

                # calling cut generators
                for cg in self.model.cut_generators:
                    cuts = cg.generate_cuts(relax_solution)
                    # adding cuts
                    for lin_expr in cuts:
                        # collecting linear expression data
                        numnz = len(lin_expr.expr)
                        cind = (c_int * numnz)()
                        cval = (c_double * numnz)()

                        # collecting variable coefficients
                        for i, (var, coeff) in enumerate(lin_expr.expr.items()):
                            cind[i] = var.idx
                            cval[i] = coeff

                        # constraint sense and rhs
                        sense = c_char(ord(lin_expr.sense))
                        rhs = c_double(-lin_expr.const)

                        GRBcbcut(p_cbdata, numnz, cind, cval, sense, rhs)

            return 0

        if self.model.cut_generators:
            self._callback = GRBcallbacktype(callback)
            GRBsetcallbackfunc(self._model, self._callback, c_void_p(0))

        # executing Gurobi to solve the formulation
        status = int(GRBoptimize(self._model))
        if status == 10009:
            raise Exception('gurobi found but license not accepted, please check it')
        
        status = c_int(0)
        st = GRBgetintattr(self._model, c_str("Status"), byref(status))
        if (st):
            raise Exception('could not check optimization status')
        
        status = status.value
        
        # todo: read solution status (code below is incomplete)
        if status == 1:  # LOADED
            return LOADED
        elif status == 2:  # OPTIMAL
            return OPTIMAL
        elif status == 3:  # INFEASIBLE
            return INFEASIBLE
        elif status == 4:  # INF_OR_UNBD
            return UNBOUNDED
        elif status == 5:  # UNBOUNDED
            return UNBOUNDED
        elif status == 6:  # CUTOFF
            return CUTOFF
        elif status == 7:  # ITERATION_LIMIT
            return -10000
        elif status == 8:  # NODE_LIMIT
            return -10000
        elif status == 9:  # TIME_LIMIT
            return -10000
        elif status == 10:  # SOLUTION_LIMIT
            return FEASIBLE
        elif status == 11:  # INTERRUPTED
            return -10000
        elif status == 12:  # NUMERIC
            return -10000
        elif status == 13:  # SUBOPTIMAL
            return FEASIBLE
        elif status == 14:  # INPROGRESS
            return -10000
        elif status == 15:  # USER_OBJ_LIMIT
            return FEASIBLE

        self._updated = True
        return status

    def get_objective_sense(self) -> str:
        isense = c_int(0)
        st = GRBgetintattr(self._model, c_str("ModelSense"), byref(isense))
        assert st == 0
        if isense.value == 1:
            return MINIMIZE
        elif isense.value == -1:
            return MAXIMIZE
        else:
            raise Exception('Unknown sense')

    def set_objective_sense(self, sense: str):
        if sense.strip().upper() == MAXIMIZE.strip().upper():
            st = GRBsetintattr(self._model, c_str("ModelSense"), c_int(-1))
            assert st == 0
        elif sense.strip().upper() == MINIMIZE.strip().upper():
            st = GRBsetintattr(self._model, c_str("ModelSense"), c_int(1))
            assert st == 0
        else:
            raise Exception("Unknown sense: {}, use {} or {}".format(sense,
                                                                     MAXIMIZE,
                                                                     MINIMIZE))

    def get_num_solutions(self) -> int:
        res = c_int(0)
        st = GRBgetintattr(self._model, c_str("SolCount"), byref(res))
        assert st == 0
        return res.value

    def var_get_xi(self, var: "Var", i: int) -> float:
        res = c_double()
        st = GRBsetintparam(GRBgetenv(self._model), c_str("SolutionNumber"),
                            c_int(i))
        assert st == 0
        st = GRBgetdblattrelement(self._model, c_str("Xn"),
                                  c_int(var.idx), byref(res))
        assert st == 0
        return res.value

    def get_objective_value_i(self, i: int) -> float:
        res = c_double(0)
        st = GRBgetdblattr(self._model, c_str("PoolObjVal"), byref(res))
        assert st == 0
        return res.value

    def get_objective_value(self) -> float:
        res = c_double(float('inf'))
        st = GRBgetdblattr(self._model, c_str('ObjVal'), byref(res))
        assert st == 0
        return res.value

    def set_processing_limits(self,
                              max_time: float = inf,
                              max_nodes: float = inf,
                              max_sol: int = inf):
        # todo: Set limits even when they are 'inf'
        if max_time != inf:
            res = GRBsetdblparam(GRBgetenv(self._model), c_str("TimeLimit"), c_double(max_time))
            assert res == 0
        if max_nodes != inf:
            res = GRBsetdblparam(GRBgetenv(self._model), c_str("NodeLimit"), c_double(max_nodes))
            assert res == 0
        if max_sol != inf:
            res = GRBsetintparam(GRBgetenv(self._model), c_str("SolutionLimit"), c_int(max_sol))
            assert res == 0

    def set_objective(self, lin_expr: "LinExpr", sense: str = "") -> None:
        # collecting linear expression data
        numnz = len(lin_expr.expr)
        cind = (c_int * numnz)()
        cval = (c_double * numnz)()

        # collecting variable coefficients
        for i, (var, coeff) in enumerate(lin_expr.expr.items()):
            cind[i] = var.idx
            cval[i] = coeff

        # objective function constant
        const = c_double(lin_expr.const)

        # resetting objective function
        num_vars = c_int(self._num_vars)
        zeros = (c_double * self._num_vars)()
        for i in range(self._num_vars):
            zeros[i] = 0.0
        GRBsetdblattrarray(self._model, c_str("Obj"), c_int(0), num_vars, zeros)

        # setting objective sense
        if sense == MAXIMIZE:
            GRBsetintattr(self._model, c_str("ModelSense"), -1)
        elif sense == MINIMIZE:
            GRBsetintattr(self._model, c_str("ModelSense"), 1)

        # setting objective function
        GRBsetdblattr(self._model, c_str("ObjCon"), const)
        GRBsetdblattrlist(self._model, c_str("Obj"), c_int(numnz), cind, cval)
        self._updated = False

    def set_objective_const(self, const: float) -> None:
        GRBsetdblattr(self._model, c_str("ObjCon"), c_double(const))

    def set_start(self, start: List[Tuple["Var", float]]) -> None:
        # collecting data
        numnz = len(start)
        cind = (c_int * numnz)()
        cval = (c_double * numnz)()

        # collecting variable coefficients
        for i in range(len(start)):
            cind[i] = start[i][0].idx
            cval[i] = start[i][1]

        GRBsetdblattrlist(self._model, c_str("Start"), numnz, cind, cval)
        self._updated = False

    def update(self) -> None:
        GRBupdatemodel(self._model)
        self._updated = True

    def write(self, file_path: str) -> None:
        # writing formulation to output file
        if not self._updated:
            self.update()
        GRBwrite(self._model, c_str(file_path))

    def read(self, file_path: str) -> None:
        GRBfreemodel(self._model)
        self._model = c_void_p(0)
        GRBreadModel(self._env, c_str(file_path), byref(self._model))

    def num_cols(self) -> int:
        res = c_int(0)
        st = GRBgetintattr(self._model, c_str('NumVars'), byref(res))
        assert st == 0
        return res.value

    def num_rows(self) -> int:
        res = c_int(0)
        st = GRBgetintattr(self._model, c_str('NumConstrs'), byref(res))
        assert st == 0
        return res.value

    def num_nz(self) -> int:
        res = c_int(0)
        st = GRBgetintattr(self._model, c_str('NumNZs'), byref(res))
        assert st == 0
        return res.value

    def get_cutoff(self) -> float:
        res = c_double(0.0)
        st = GRBgetdblparam(GRBgetenv(self._model), c_str("Cutoff"), byref(res))
        assert st == 0
        return res.value

    def set_cutoff(self, cutoff: float):
        st = GRBsetdblparam(GRBgetenv(self._model), c_str("Cutoff"), c_double(cutoff))
        assert st == 0

    def get_mip_gap_abs(self) -> float:
        res = c_double(0.0)
        st = GRBgetdblparam(GRBgetenv(self._model), c_str("MIPGapAbs"), byref(res))
        assert st == 0
        return res.value

    def set_mip_gap_abs(self, allowable_gap: float):
        st = GRBsetdblparam(GRBgetenv(self._model), c_str("MIPGapAbs"),
                            c_double(allowable_gap))
        assert st == 0

    def get_mip_gap(self) -> float:
        res = c_double(0.0)
        st = GRBgetdblparam(GRBgetenv(self._model), c_str("MIPGap"), byref(res))
        assert st == 0
        return res.value

    def set_mip_gap(self, allowable_ratio_gap: float):
        st = GRBsetdblparam(GRBgetenv(self._model), c_str("MIPGap"),
                            c_double(allowable_ratio_gap))
        assert st == 0

    def constr_get_expr(self, constr: Constr) -> LinExpr:
        if not self._updated:
            self.update()

        numnz = c_int()
        cbeg = POINTER(c_int)()
        cind = POINTER(c_int)()
        cval = POINTER(c_double)()

        # obtaining number of non-zeros
        GRBgetconstrs(self._model, byref(numnz), cbeg, cind, cval, c_int(constr.idx), c_int(1))

        # creating arrays to hold indices and coefficients
        cbeg = (c_int * 2)()  # beginning and ending
        cind = (c_int * numnz.value)()
        cval = (c_double * numnz.value)()

        # obtaining variables and coefficients
        GRBgetconstrs(self._model, byref(numnz), cbeg, cind, cval, c_int(constr.idx), c_int(1))

        # obtaining sense and rhs
        c_sense = c_char()
        rhs = c_double()
        GRBgetcharattrelement(self._model, c_str("Sense"), c_int(constr.idx), byref(c_sense))
        GRBgetdblattrelement(self._model, c_str("RHS"), c_int(constr.idx), byref(rhs))

        # translating sense
        sense = ""
        if c_sense.value == b"<":
            sense = LESS_OR_EQUAL
        elif c_sense.value == b">":
            sense = GREATER_OR_EQUAL
        elif c_sense.value == b"=":
            sense = EQUAL

        expr = LinExpr(const=-rhs.value, sense=sense)
        for i in range(numnz.value):
            expr.add_var(self.model.vars[cind[i]], cval[i])

        return expr

    def constr_get_name(self, idx: int) -> str:
        vName = c_char_p(0)
        st = GRBgetstrattrelement(self._model, c_str('ConstrName'), c_int(idx), byref(vName))
        assert st == 0
        return vName.value.decode('utf-8')

    def constr_set_expr(self, constr: Constr, value: LinExpr) -> LinExpr:
        raise NotImplementedError("Gurobi: functionality currently unavailable via PyMIP...")

    def constr_get_pi(self, constr: "Constr") -> float:
        res = c_double()
        st = GRBgetdblattrelement(self._model, c_str("Pi"), c_int(constr.idx), byref(res))
        assert st == 0
        return res.value

    def var_get_lb(self, var: "Var") -> float:
        if not self._updated:
            self.update()

        res = c_double()
        st = GRBgetdblattrelement(self._model, c_str("LB"), c_int(var.idx), byref(res))
        assert st == 0
        return res.value

    def var_set_lb(self, var: "Var", value: float) -> None:
        GRBsetdblattrelement(self._model, c_str("LB"), c_int(var.idx), c_double(value))
        self._updated = False

    def var_get_ub(self, var: "Var") -> float:
        if not self._updated:
            self.update()

        res = c_double()
        GRBgetdblattrelement(self._model, c_str("UB"), c_int(var.idx), byref(res))
        return res.value

    def var_set_ub(self, var: "Var", value: float) -> None:
        GRBsetdblattrelement(self._model, c_str("UB"), c_int(var.idx), c_double(value))
        self._updated = False

    def var_get_obj(self, var: "Var") -> float:
        if not self._updated:
            self.update()

        res = c_double()
        GRBgetdblattrelement(self._model, c_str("Obj"), c_int(var.idx), byref(res))
        return res.value

    def var_set_obj(self, var: "Var", value: float) -> None:
        GRBsetdblattrelement(self._model, c_str("Obj"), c_int(var.idx), c_double(value))
        self._updated = False

    def var_get_type(self, var: "Var") -> str:
        if not self._updated:
            self.update()

        res = c_char(0)
        GRBgetcharattrelement(self._model, c_str("VType"), c_int(var.idx), byref(res))

        if res.value == b"B":
            return BINARY
        elif res.value == b"C":
            return CONTINUOUS
        elif res.value == b"I":
            return INTEGER

        raise ValueError("Gurobi: invalid variable type returned...")

    def var_set_type(self, var: "Var", value: str) -> None:
        if value == BINARY:
            vtype = c_char(ord("B"))
        elif value == CONTINUOUS:
            vtype = c_char(ord("C"))
        elif value == INTEGER:
            vtype = c_char(ord("I"))
        else:
            raise ValueError("Gurobi: invalid variable type...")

        GRBsetcharattrelement(self._model, c_str("VType"), c_int(var.idx), vtype)
        self._updated = False

    def var_get_column(self, var: "Var"):
        if not self._updated:
            self.update()

        numnz = c_int()
        cbeg = POINTER(c_int)()
        cind = POINTER(c_int)()
        cval = POINTER(c_double)()

        # obtaining number of non-zeros
        GRBgetvars(self._model, byref(numnz), cbeg, cind, cval, c_int(var.idx), c_int(1))

        # creating arrays to hold indices and coefficients
        cbeg = (c_int * 2)()  # beginning and ending
        cind = (c_int * numnz.value)()
        cval = (c_double * numnz.value)()

        # obtaining variables and coefficients
        GRBgetvars(self._model, byref(numnz), cbeg, cind, cval, c_int(var.idx), c_int(1))

        constr = [self.model.constrs[cind[i]] for i in range(numnz.value)]
        coefs = [float(cval[i]) for i in range(numnz.value)]

        col = Column(constr, coefs)

        return col

    def var_set_column(self, var: "Var", value: Column):
        raise NotImplementedError("Gurobi: functionality currently unavailable via PyMIP...")

    def var_get_rc(self, var: "Var") -> float:
        res = c_double()
        GRBgetdblattrelement(self._model, c_str("RC"), c_int(var.idx), byref(res))
        return res.value

    def var_get_x(self, var: Var) -> float:
        res = c_double()
        GRBgetdblattrelement(self._model, c_str("X"), c_int(var.idx), byref(res))
        return res.value

    def var_get_name(self, idx: int) -> str:
        vName = c_char_p(0)
        st = GRBgetstrattrelement(self._model, c_str('VarName'), c_int(idx), byref(vName))
        assert st == 0
        return vName.value.decode('utf-8')

    def get_emphasis(self) -> int:
        fc = c_int(0)
        st = GRBgetintparam(GRBgetenv(self._model), c_str("MIPFocus"),
                            byref(fc))
        assert st == 0
        if fc == 1:
            return FEASIBILITY
        elif fc == 3 or fc == 2:
            return OPTIMALITY

        return 0

    def set_emphasis(self, emph: int):
        if emph == FEASIBILITY:
            st = GRBsetintparam(GRBgetenv(self._model), c_str("MIPFocus"),
                                c_int(1))
            assert st == 0
        elif emph == OPTIMALITY:
            st = GRBsetintparam(GRBgetenv(self._model), c_str("MIPFocus"),
                                c_int(2))
            assert st == 0
        else:
            st = GRBsetintparam(GRBgetenv(self._model), c_str("MIPFocus"),
                                c_int(0))
            assert st == 0


# auxiliary functions
def c_str(value) -> c_char_p:
    """
    This function converts a python string into a C compatible char[]
    :param value: input string
    :return: string converted to C"s format
    """
    return create_string_buffer(value.encode("utf-8"))


has_gurobi = False

try:
    found = False
    libPath = None

    for majorVersion in reversed(range(2, 12)):
        for minorVersion in reversed(range(0, 11)):
            try:
                libPath = find_library('gurobi{}{}'.format(majorVersion, minorVersion))
                if libPath != None:
                    break
            except:
                continue
        if libPath != None:
            break

    if libPath == None:
        raise Exception()
    grblib = CDLL(libPath)
    print('gurobi version {}.{} found'.format(majorVersion, minorVersion))
    has_gurobi = True
except:
    has_gurobi = False
# create/release environment and model

if has_gurobi:
    GRBloadenv = grblib.GRBloadenv
    GRBloadenv.restype = c_int
    GRBloadenv.argtypes = [c_void_p, c_char_p]

    GRBnewmodel = grblib.GRBnewmodel
    GRBnewmodel.restype = c_int
    GRBnewmodel.argtypes = [c_void_p, c_void_p, c_char_p, c_int, POINTER(c_double), POINTER(c_double),
                            POINTER(c_double), c_char_p, c_void_p]

    GRBfreeenv = grblib.GRBfreeenv
    GRBfreeenv.restype = c_int
    GRBfreeenv.argtypes = [c_void_p]

    GRBfreemodel = grblib.GRBfreemodel
    GRBfreemodel.argtypes = [c_void_p]

    # manipulate attributes

    GRBgetintattr = grblib.GRBgetintattr
    GRBgetintattr.restype = c_int
    GRBgetintattr.argtypes = [c_void_p, c_char_p, POINTER(c_int)]

    GRBsetintattr = grblib.GRBsetintattr
    GRBsetintattr.restype = c_int
    GRBsetintattr.argtypes = [c_void_p, c_char_p, c_int]

    GRBgetintattrelement = grblib.GRBgetintattrelement
    GRBgetintattrelement.restype = c_int
    GRBgetintattrelement.argtypes = [c_void_p, c_char_p, c_int, POINTER(c_int)]

    GRBsetintattrelement = grblib.GRBsetintattrelement
    GRBsetintattrelement.restype = c_int
    GRBsetintattrelement.argtypes = [c_void_p, c_char_p, c_int, c_int]

    GRBgetdblattr = grblib.GRBgetdblattr
    GRBgetdblattr.restype = c_int
    GRBgetdblattr.argtypes = [c_void_p, c_char_p, POINTER(c_double)]

    GRBsetdblattr = grblib.GRBsetdblattr
    GRBsetdblattr.restype = c_int
    GRBsetdblattr.argtypes = [c_void_p, c_char_p, c_double]

    GRBgetdblattrarray = grblib.GRBgetdblattrarray
    GRBgetdblattrarray.restype = c_int
    GRBgetdblattrarray.argtypes = [c_void_p, c_char_p, c_int, c_int, POINTER(c_double)]

    GRBsetdblattrarray = grblib.GRBsetdblattrarray
    GRBsetdblattrarray.restype = c_int
    GRBsetdblattrarray.argtypes = [c_void_p, c_char_p, c_int, c_int, POINTER(c_double)]

    GRBsetdblattrlist = grblib.GRBsetdblattrlist
    GRBsetdblattrlist.restype = c_int
    GRBsetdblattrlist.argtypes = [c_void_p, c_char_p, c_int, POINTER(c_int), POINTER(c_double)]

    GRBgetdblattrelement = grblib.GRBgetdblattrelement
    GRBgetdblattrelement.restype = c_int
    GRBgetdblattrelement.argtypes = [c_void_p, c_char_p, c_int, POINTER(c_double)]

    GRBsetdblattrelement = grblib.GRBsetdblattrelement
    GRBsetdblattrelement.restype = c_int
    GRBsetdblattrelement.argtypes = [c_void_p, c_char_p, c_int, c_double]

    GRBsetcharattrarray = grblib.GRBsetcharattrarray
    GRBsetcharattrarray.restype = c_int
    GRBsetcharattrarray.argtypes = [c_void_p, c_char_p, c_int, c_int, c_char_p]

    GRBgetcharattrelement = grblib.GRBgetcharattrelement
    GRBgetcharattrelement.restype = c_int
    GRBgetcharattrelement.argtypes = [c_void_p, c_char_p, c_int, POINTER(c_char)]

    GRBsetcharattrelement = grblib.GRBsetcharattrelement
    GRBsetcharattrelement.restype = c_int
    GRBsetcharattrelement.argtypes = [c_void_p, c_char_p, c_int, c_char]

    GRBgetstrattrelement = grblib.GRBgetstrattrelement
    GRBgetstrattrelement.argtypes = [c_void_p, c_char_p, c_int, POINTER(c_char_p)]
    GRBgetstrattrelement.restype = c_int

    # manipulate parameter(s)

    GRBgetintparam = grblib.GRBgetintparam
    GRBgetintparam.argtypes = [c_void_p, c_char_p, POINTER(c_int)]
    GRBgetintparam.restype = c_int

    GRBsetintparam = grblib.GRBsetintparam
    GRBsetintparam.argtypes = [c_void_p, c_char_p, c_int]
    GRBsetintparam.restype = c_int

    GRBgetdblparam = grblib.GRBgetdblparam
    GRBgetdblparam.argtypes = [c_void_p, c_char_p, POINTER(c_double)]
    GRBgetdblparam.restype = c_int

    GRBsetdblparam = grblib.GRBsetdblparam
    GRBsetdblparam.argtypes = [c_void_p, c_char_p, c_double]
    GRBsetdblparam.restype = c_int

    # manipulate objective function(s)

    GRBsetobjectiven = grblib.GRBsetobjectiven
    GRBsetobjectiven.restype = c_int
    GRBsetobjectiven.argtypes = [c_void_p, c_int, c_int, c_double, c_double, c_double, c_char_p,
                                 c_double, c_int, POINTER(c_int), POINTER(c_double)]

    # add variables and constraints

    GRBaddvar = grblib.GRBaddvar
    GRBaddvar.restype = c_int
    GRBaddvar.argtypes = [c_void_p, c_int, POINTER(c_int), POINTER(c_double), c_double, c_double,
                          c_double, c_char, c_char_p]

    GRBaddconstr = grblib.GRBaddconstr
    GRBaddconstr.restype = c_int
    GRBaddconstr.argtypes = [c_void_p, c_int, POINTER(c_int), POINTER(c_double), c_char, c_double,
                             c_char_p]

    # get constraints

    GRBgetconstrs = grblib.GRBgetconstrs
    GRBgetconstrs.restype = c_int
    GRBgetconstrs.argtypes = [c_void_p, POINTER(c_int), POINTER(c_int), POINTER(c_int),
                              POINTER(c_double), c_int, c_int]

    # get variables
    GRBgetvars = grblib.GRBgetvars
    GRBgetvars.argtypes = [c_void_p, POINTER(c_int), POINTER(c_int),
                           POINTER(c_int), POINTER(c_double), c_int, c_int]
    GRBgetvars.restype = c_int

    # callback functions and constants

    GRBcallbacktype = CFUNCTYPE(c_int, c_void_p, c_void_p, c_int, c_void_p)

    GRBsetcallbackfunc = grblib.GRBsetcallbackfunc
    GRBsetcallbackfunc.restype = c_int
    GRBsetcallbackfunc.argtypes = [c_void_p, GRBcallbacktype, c_void_p]

    GRBcbcut = grblib.GRBcbcut
    GRBcbcut.restype = c_int
    GRBcbcut.argtypes = [c_void_p, c_int, POINTER(c_int), POINTER(c_double), c_char, c_double]

    GRBcbget = grblib.GRBcbget
    GRBcbget.restype = c_int
    GRBcbget.argtypes = [c_void_p, c_int, c_int, c_void_p]

    GRBcbsolution = grblib.GRBcbsolution
    GRBcbsolution.restype = c_int
    GRBcbsolution.argtypes = [c_void_p, POINTER(c_double), POINTER(c_double)]

    GRB_CB_PRE_COLDEL = 1000
    GRB_CB_PRE_ROWDEL = 1001
    GRB_CB_PRE_SENCHG = 1002
    GRB_CB_PRE_BNDCHG = 1003
    GRB_CB_PRE_COECHG = 1004

    GRB_CB_SPX_ITRCNT = 2000
    GRB_CB_SPX_OBJVAL = 2001
    GRB_CB_SPX_PRIMINF = 2002
    GRB_CB_SPX_DUALINF = 2003
    GRB_CB_SPX_ISPERT = 2004

    GRB_CB_MIP_OBJBST = 3000
    GRB_CB_MIP_OBJBND = 3001
    GRB_CB_MIP_NODCNT = 3002
    GRB_CB_MIP_SOLCNT = 3003
    GRB_CB_MIP_CUTCNT = 3004
    GRB_CB_MIP_NODLFT = 3005
    GRB_CB_MIP_ITRCNT = 3006

    GRB_CB_MIPSOL_SOL = 4001
    GRB_CB_MIPSOL_OBJ = 4002
    GRB_CB_MIPSOL_OBJBST = 4003
    GRB_CB_MIPSOL_OBJBND = 4004
    GRB_CB_MIPSOL_NODCNT = 4005
    GRB_CB_MIPSOL_SOLCNT = 4006

    GRB_CB_MIPNODE_STATUS = 5001
    GRB_CB_MIPNODE_REL = 5002
    GRB_CB_MIPNODE_OBJBST = 5003
    GRB_CB_MIPNODE_OBJBND = 5004
    GRB_CB_MIPNODE_NODCNT = 5005
    GRB_CB_MIPNODE_SOLCNT = 5006

    GRB_CB_MSG_STRING = 6001
    GRB_CB_RUNTIME = 6002

    # optimize/update model

    GRBoptimize = grblib.GRBoptimize
    GRBoptimize.restype = c_int
    GRBoptimize.argtypes = [c_void_p]

    GRBupdatemodel = grblib.GRBupdatemodel
    GRBupdatemodel.restype = c_int
    GRBupdatemodel.argtypes = [c_void_p]

    # read/write files

    GRBwrite = grblib.GRBwrite
    GRBwrite.restype = c_int
    GRBwrite.argtypes = [c_void_p, c_char_p]

    GRBreadModel = grblib.GRBreadmodel
    GRBreadModel.restype = c_int
    GRBreadModel.argtypes = [c_void_p, c_char_p, c_void_p]

    GRBgetenv = grblib.GRBgetenv
    GRBgetenv.restype = c_void_p
    GRBgetenv.argtypes = [c_void_p]

# vim: ts=4 sw=4 et
